# -*- coding: utf-8 -*-
"""
Created on Wed Mar 11 18:00:28 2020

@author: gamelina
"""

import at
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d


class Optics:
    """Handle optic functions"""
    def __init__(self, lattice_file=None, local_beta=None, local_alpha=None, 
                 local_dispersion=None, **kwargs):
        
        if lattice_file is not None:
            self.use_local_values = False
            self.load_from_AT(lattice_file, **kwargs)
            if local_beta is None:
                self.local_beta = np.mean(self.beta_array)
            else:
                self.local_beta = local_beta
            if local_alpha is None:
                self.local_alpha = np.mean(self.alpha_array)
            else:
                self.local_alpha = local_alpha
            if local_dispersion is None:
                self.local_dispersion = np.zeros((4,))
            else:
                self.local_dispersion = local_dispersion
            self.local_gamma = (1 + self.local_alpha**2)/self.local_beta
            
        else:
            self.use_local_values = True
            self.local_beta = local_beta
            self.local_alpha = local_alpha
            self.local_gamma = (1 + self.local_alpha**2)/self.local_beta
            self.local_dispersion = local_dispersion

    def load_from_AT(self, lattice_file, **kwargs):
        
        self.n_points = int(kwargs.get("n_points", 1e3))
        periodicity = kwargs.get("periodicity")
        
        self.lattice = at.load_lattice(lattice_file)
        lattice = self.lattice.slice(slices=self.n_points)
        refpts = np.arange(0, len(lattice))
        twiss0, tune, chrom, twiss = at.get_twiss(lattice, refpts=refpts,
                                                  get_chrom=True)
        
        if periodicity is None:
            self.periodicity = lattice.periodicity
        else:
            self.periodicity = periodicity
                    
        for i in range(self.periodicity-1):
            pos = np.append(twiss.s_pos, twiss.s_pos + twiss.s_pos[-1]*(i+1))
            
        self.position = pos
        self.beta_array = np.tile(twiss.beta.T, self.periodicity)
        self.alpha_array = np.tile(twiss.alpha.T, self.periodicity)
        self.gamma_array = (1 + self.alpha_array**2)/self.beta_array
        self.dispersion_array = np.tile(twiss.dispersion.T, self.periodicity)
        self.tune = tune * self.periodicity
        self.chro = chrom * self.periodicity
        self.ac = at.get_mcf(self.lattice)
        
        self.setup_interpolation()
        
        
    def setup_interpolation(self):        
        self.betaX = interp1d(self.position, self.beta_array[0,:],
                              kind='linear')
        self.betaY = interp1d(self.position, self.beta_array[1,:],
                              kind='linear')
        self.alphaX = interp1d(self.position, self.alpha_array[0,:],
                               kind='linear')
        self.alphaY = interp1d(self.position, self.alpha_array[1,:],
                               kind='linear')
        self.gammaX = interp1d(self.position, self.gamma_array[0,:],
                               kind='linear')
        self.gammaY = interp1d(self.position, self.gamma_array[1,:],
                               kind='linear')
        self.dispX = interp1d(self.position, self.dispersion_array[0,:],
                              kind='linear')
        self.disppX = interp1d(self.position, self.dispersion_array[1,:],
                               kind='linear')
        self.dispY = interp1d(self.position, self.dispersion_array[2,:],
                              kind='linear')
        self.disppY = interp1d(self.position, self.dispersion_array[3,:],
                               kind='linear')
        
    def beta(self, position):
        if self.use_local_values:
            return np.outer(self.local_beta, np.ones_like(position))
        else:
            beta = [self.betaX(position), self.betaY(position)]
            return np.array(beta)
    
    def alpha(self, position):
        if self.use_local_values:
            return np.outer(self.local_alpha, np.ones_like(position))
        else:
            alpha = [self.alphaX(position), self.alphaY(position)]
            return np.array(alpha)
    
    def gamma(self, position):
        if self.use_local_values:
            return np.outer(self.local_gamma, np.ones_like(position))
        else:
            gamma = [self.gammaX(position), self.gammaY(position)]
            return np.array(gamma)
    
    def dispersion(self, position):
        if self.use_local_values:
            return np.outer(self.local_dispersion, np.ones_like(position))
        else:
            dispersion = [self.dispX(position), self.disppX(position), 
                          self.dispY(position), self.disppY(position)]
            return np.array(dispersion)
    
class PhyisicalModel:
    """Store lattice physical parameters"""
    def __init__(self, ring, a0=5e-3, b0=5e-3, shape='circ', rho=1.7e-8,
                 n_points=1e4):
        
        self.n_points = int(n_points)
        self.position = np.linspace(0, ring.L, self.n_points)
        self.x0 = np.ones_like(self.position)*a0 # Horizontal half aperture [m]
        self.y0 = np.ones_like(self.position)*b0 # Vertical half aperture [m]
        
        self.length = self.position[1:] - self.position[:-1]
        self.center = (self.position[1:] + self.position[:-1])/2
        self.rho = np.ones_like(self.center)*rho # Resistivity of the chamber material [ohm*m]
        
        self.shape = np.repeat(np.array([shape]), self.n_points-1) # Shape of the chamber cross section (elli/circ/rect)

    def resistive_wall_coef(self):
        L = self.end_position[-1]
        sigma = 1/self.rho
        beta_H_star = 1/L*(self.length*self.beta[0,:]).sum()
        beta_V_star = 1/L*(self.length*self.beta[1,:]).sum()
        sigma_star = 1/L*(self.length*sigma).sum()
        
        a1_H = (((self.length*self.beta[0,:]/(np.sqrt(sigma)*(self.a0)**1)).sum())**(-1)*L*beta_H_star/np.sqrt(sigma_star))**(1/1)
        a2_H = (((self.length*self.beta[0,:]/(np.sqrt(sigma)*(self.a0)**2)).sum())**(-1)*L*beta_H_star/np.sqrt(sigma_star))**(1/2)

        a1_V = (((self.length*self.beta[1,:]/(np.sqrt(sigma)*(self.b0)**1)).sum())**(-1)*L*beta_V_star/np.sqrt(sigma_star))**(1/1)
        a2_V = (((self.length*self.beta[1,:]/(np.sqrt(sigma)*(self.b0)**2)).sum())**(-1)*L*beta_V_star/np.sqrt(sigma_star))**(1/2)
        
        a3_H = (((self.length*self.beta[0,:]/(np.sqrt(sigma)*(self.a0)**3)).sum())**(-1)*L*beta_H_star/np.sqrt(sigma_star))**(1/3)
        a4_H = (((self.length*self.beta[0,:]/(np.sqrt(sigma)*(self.a0)**4)).sum())**(-1)*L*beta_H_star/np.sqrt(sigma_star))**(1/4)
        
        a3_V = (((self.length*self.beta[1,:]/(np.sqrt(sigma)*(self.b0)**3)).sum())**(-1)*L*beta_V_star/np.sqrt(sigma_star))**(1/3)
        a4_V = (((self.length*self.beta[1,:]/(np.sqrt(sigma)*(self.b0)**4)).sum())**(-1)*L*beta_V_star/np.sqrt(sigma_star))**(1/4)
        
        a1_L = min((a1_H,a1_V))
        a2_L = min((a2_H,a2_V))
        
        print("rhorw = " + str(1/sigma_star))
        print("beffL1 = " + str(a1_L))
        print("beffL2 = " + str(a2_L))
        print("beffH3 = " + str(a3_H))
        print("beffH4 = " + str(a4_H))
        print("beffV3 = " + str(a3_V))
        print("beffV4 = " + str(a4_V))
        
    def change_values(self, pos1, pos2, a0, b0, shape, rho):
        ind = (self.position > pos1) & (self.position < pos2)
        self.x0[ind] = a0
        self.y0[ind] = b0
        ind2 = (self.position[:-1] > pos1) & (self.position[1:] < pos2)
        self.rho[ind2] = rho
        self.shape[ind2] = shape
        
    def taper(self, pos1, pos2, x0_start, y0_start, x0_end, y0_end, shape, rho):
        ind = (self.position > pos1) & (self.position < pos2)
        self.x0[ind] = np.linspace(x0_start, x0_end, sum(ind))
        self.y0[ind] = np.linspace(y0_start, y0_end, sum(ind))
        ind2 = (self.position[:-1] > pos1) & (self.position[1:] < pos2)
        self.rho[ind2] = rho
        self.shape[ind2] = shape
        
    def plot_aperture(self):
        plt.plot(self.position,self.x0*1e3)
        plt.plot(self.position,self.y0*1e3)
        plt.xlabel("Longitudinal position [m]")
        plt.ylabel("H/V aperture [mm]")
        plt.legend(["H","V"])
